/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <string.h>
#include <ctype.h>

#include <gnome.h>

#include "article.h"
#include "article-db.h"
#include "articlelist.h"
#include "article-thread.h"
#include "fnmatch.h"
#include "group.h"
#include "pan.h"
#include "queue.h"
#include "status-item.h"
#include "queue-item-body.h"
#include "queue-item-decode.h"
#include "queue-item-headers.h"
#include "text.h"

#include "xpm/book_open.xpm"
#include "xpm/book_closed.xpm"
#include "xpm/mini_page.xpm"
#include "xpm/greencheck.xpm"
#include "xpm/bluecheck.xpm"
#include "xpm/thread.xpm"
#include "xpm/x.xpm"
#include "xpm/binary.xpm"
#include "xpm/binary_complete.xpm"
#include "xpm/binary_incomplete.xpm"
#include "xpm/unread_children.xpm"


/**********************************************************
 * Private Variables
 **********************************************************/

/* state */
static group_data *my_group = NULL;
static server_data *my_server = NULL;
static article_data** article_headers = NULL;
static int article_header_qty = 0;
static gboolean is_threaded = TRUE;
static int sort_type = ARTICLE_SORT_SUBJECT;
static GHashTable *sort_type_to_menu_index = NULL;
static GtkWidget *sort_option_menu = NULL;
static gchar* subject_filter_str = NULL;

/* by holding a reference to the active group, we can avoid a lot of open/close
 * db calls as the user reads articles.  We don't * actually do anything with
 * it here, but we know that most of the article headers being updated in the
 * db will be through the article list. */
static article_db active_group_header_db = NULL;

static void sort_selected_cb (GtkWidget* w, gpointer data);

/* search */
static GtkCTreeNode* find_article_next_pos = NULL;
static GtkWidget *find_dialog;

/* article list */
static GnomePixmap *articlelist_closed_pixmap = NULL;
static GnomePixmap *articlelist_open_pixmap = NULL;
static GnomePixmap *decode_queue_pixmap = NULL;
static GnomePixmap *download_queue_pixmap = NULL;
static GnomePixmap *mp_complete_pixmap = NULL;
static GnomePixmap *mp_incomplete_pixmap = NULL;
static GnomePixmap *error_pixmap = NULL;
static GnomePixmap *article_read_pixmap = NULL;
static GnomePixmap *binary_decoded_pixmap = NULL;
static GnomePixmap *unread_children_pixmap = NULL;

/**********************************************************
 * Public Variables
 **********************************************************/

GdkColormap *cmap;
GdkColor new_color;
GdkColor read_color;
GdkColor unread_color;
GdkColor killfile_color;

article_data *current_article = NULL;
GtkWidget *article_ctree_menu;

GtkStyle *killfile_style = NULL;
GtkStyle *unread_style = NULL;
GtkStyle *read_style = NULL;
GtkStyle *new_style = NULL;

gboolean hide_mpart_child_nodes = TRUE;

/**********************************************************
 * Private Functions
 **********************************************************/

/* search */
static void find_article (const gchar*, const gchar*, gboolean, gboolean );
static void find_dialog_clicked_cb (GnomeDialog *dialog, int button, GtkWidget *find_entry);

/* ui */
static GtkCTreeNode* articlelist_get_node_from_message_id ( const char* message_id, int* index );
static void articlelist_update (const group_data*, article_data**, int qty, StatusItem*);
static void build_article_ctree_recursive (GtkCTree*, GtkCTreeNode*, GSList*, StatusItem*, int);
static GnomePixmap* get_article_pixmap (const article_data*);
static void articlelist_button_press (GtkWidget*, GdkEventButton*);
static void articlelist_click_column (GtkCList *, int col_clicked);
static void article_ctree_destroy_cb ( void );
static void articlelist_update_node_fast ( GtkCTreeNode*, const article_data* );



/**********************************************************
 * Articlelist popup menu
 **********************************************************/

static GnomeUIInfo articlelist_menu_popup[] = {
	{ GNOME_APP_UI_ITEM,
		N_("Open in Window"),
		N_("Open this message for reading."),
		message_read_window },
#if 0
	{ GNOME_APP_UI_ITEM,
		N_("Print"),
		N_("Print this message out to a physical medium."),
		not_implemented },
#endif
	GNOMEUIINFO_SEPARATOR,
        { GNOME_APP_UI_ITEM,
		N_("_Followup to newsgroup"),
		N_("Post a reply to the message on the news server."),
		message_followup_window },
	{ GNOME_APP_UI_ITEM,
		N_("Reply by _E-mail"),
		N_("Create a mail reply to the sender."),
		message_reply_window },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM,
		N_("Save Attachment"),
		N_("Save Attachments."),
		articlelist_selected_decode,
	  	NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL,
		'D', GDK_CONTROL_MASK, NULL },
	{ GNOME_APP_UI_ITEM,
		N_("Open Attachment"),
		N_("Open Attachments."),
		articlelist_selected_open,
		NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL,
		'O', GDK_CONTROL_MASK, NULL },
	{ GNOME_APP_UI_ITEM,
		N_("Delete Message"),
		N_("Delete message from database."),
		articlelist_selected_delete },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM,
		N_("Mark as Read"),
		N_("Marks current message as read."),
		articlelist_selected_mark_read },
	{ GNOME_APP_UI_ITEM,
		N_("Mark as Unread"),
		N_("Marks the current message as unread."),
		articlelist_selected_mark_unread },
	GNOMEUIINFO_SEPARATOR,
        { GNOME_APP_UI_ITEM,
		N_("Mark All Articles Read"),
		N_("Marks the entire group as read."),
		articlelist_all_mark_read },
	GNOMEUIINFO_END
};

static const char*
articlelist_get_row_message_id (int i)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* node = gtk_ctree_node_nth (tree, i);
	const article_data* adata = NULL;
	const gchar* message_id = NULL;

	if (node)
		adata = gtk_ctree_node_get_row_data (tree, node);
	if (adata)
		message_id = adata->message_id;

	return message_id;
}

static GtkCTreeNode*
articlelist_get_selected_node (void)
{
	GtkCTreeNode* node = NULL;

	const GList* list =  GTK_CLIST(Pan.article_ctree)->selection;
	if ( list != NULL )
		node = GTK_CTREE_NODE(list->data);

	return node;
}

static void
articlelist_unselect_all (void)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);

	pan_lock();
	gtk_clist_freeze (GTK_CLIST(tree));
	gtk_ctree_unselect_recursive (tree, gtk_ctree_node_nth(tree, 0));
	gtk_clist_thaw (GTK_CLIST(tree));
	pan_unlock();
}

static void
articlelist_set_selected_node (GtkCTreeNode *node)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* n = NULL;

	pan_lock();
	gtk_clist_freeze (GTK_CLIST(tree));
	g_assert ( node!=NULL );
	gtk_ctree_unselect_recursive (tree, gtk_ctree_node_nth(tree, 0));
	for (n=GTK_CTREE_ROW(node)->parent; n; n=GTK_CTREE_ROW(n)->parent)
		gtk_ctree_expand(tree, n);
	gtk_ctree_select (tree, node);
	gtk_clist_thaw (GTK_CLIST(tree));
	gtk_ctree_node_moveto (tree, node, 0, 0.5, 0);
	pan_unlock();
}

static void
articlelist_select_and_read_node (GtkCTreeNode *node)
{
	GtkCTree *tree = GTK_CTREE(Pan.article_ctree);
	article_data *adata = NULL;

	if (node!=NULL)
		adata = (article_data*) gtk_ctree_node_get_row_data(tree,node);
	if (adata!=NULL)
	{
		articlelist_set_selected_node(node);
		queue_add(QUEUE_ITEM(queue_item_body_new(
			my_server, my_group, message_new(my_server, my_group, adata))));
	}
}

void
articlelist_set_selected_message_id (const char* message_id)
{
	GtkCTreeNode* node = articlelist_get_node_from_message_id (message_id, NULL);
	g_return_if_fail (node!=NULL);
	articlelist_set_selected_node (node);
}

const char*
articlelist_get_selected_message_id (void)
{
	GtkCTreeNode* node = articlelist_get_selected_node ();
	const article_data* adata = (const article_data*)
		gtk_ctree_node_get_row_data (GTK_CTREE(Pan.article_ctree), node);
	return adata ? adata->message_id : NULL;
}

static GtkCTreeNode*
articlelist_node_next (GtkCTreeNode *node, gboolean sibling_only)
{
	GtkCTreeNode* n;

	if (!node)
		return NULL;

	if (sibling_only)
		return GTK_CTREE_ROW(node)->sibling;

	if (GTK_CTREE_ROW(node)->children)
		return GTK_CTREE_ROW(node)->children;

	for (n=node; n!=NULL; n=GTK_CTREE_ROW(n)->parent)
		if (GTK_CTREE_ROW(n)->sibling)
			return GTK_CTREE_ROW(n)->sibling;

	return NULL;
}


static GtkCTreeNode*
articlelist_node_prev (GtkCTreeNode *node, gboolean sibling_only)
{
	GtkCTreeNode *parent;
	GtkCTreeNode *sibling;
	
	if (!node)
		return NULL;

	/* get parent */	
	parent = GTK_CTREE_ROW(node)->parent;
	if (!parent)
		return NULL;

	/* parent's first child */
	sibling=GTK_CTREE_ROW(parent)->children;
	if (sibling==node) /* this is the first child */
		return parent;

	/* previous sibling of node */
	while (GTK_CTREE_ROW(sibling)->sibling != node) {
		sibling = GTK_CTREE_ROW(sibling)->sibling;
		g_assert (sibling != NULL);
	}

	if (sibling_only)
		return sibling;

	/* find the absolutely last child of the older sibling */
	for (;;) {
		GtkCTreeNode* tmp = GTK_CTREE_ROW(sibling)->children;
		if (!tmp)
			return sibling;

		/* last child of sibling */
		while (GTK_CTREE_ROW(tmp)->sibling != NULL)
			tmp = GTK_CTREE_ROW(tmp)->sibling;

		sibling = tmp;
	}

	pan_warn_if_reached ();
}

static const article_data*
articlelist_get_relative_article (const char* message_id, int step)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode *node=NULL;
	const int row_qty = GTK_CLIST(tree)->rows;

	if ( row_qty<2 )
		return NULL;

	/* get working node */
	node = articlelist_get_node_from_message_id (message_id, NULL);
	g_assert ( node!=NULL );

	/* Find the requested node... */
	while ( step > 0 ) {
		--step;
		node = articlelist_node_next (node, FALSE);
	}
	while ( step < 0 ) {
		++step;
		node = articlelist_node_prev (node, FALSE);
	}


	/* get step node */
	return (const article_data*) gtk_ctree_node_get_row_data (tree, node);
}


const article_data*
articlelist_get_next_article (void)
{
	const article_data *adata = NULL;

	const char* message_id = articlelist_get_selected_message_id ();
	if (!message_id)
		message_id = articlelist_get_row_message_id (1);
	if (message_id)
		adata = articlelist_get_relative_article (message_id, 1);

	return adata;
}

void
articlelist_read_next_article(void)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* node;

	/* if nothing to read... */
	if ( !GTK_CLIST(tree)->rows )
		return;

	/* get starting point */
	node = articlelist_get_selected_node ();
	if (!node)
		node = gtk_ctree_node_nth (tree, 0);

	/* move to next node */
	node = articlelist_node_next (node, FALSE);

	/* read it */
	if ( node!=NULL )
		articlelist_select_and_read_node (node);
	else {
		articlelist_unselect_all ();
		articlelist_read_next_article();
	}
}
void
articlelist_read_next_unread_article (void)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* node = NULL;
	GtkCTreeNode* start = NULL;
	gboolean all_read = FALSE;

	/* get first node */	
	node = articlelist_get_selected_node ();
	if (!node)
		node = gtk_ctree_node_nth (tree, 0);

	/* move to next unread node */
	all_read = FALSE;
	start = node;
	for ( ;; ) {
		const article_data* adata = (const article_data*)
			gtk_ctree_node_get_row_data (tree, node);
		if (adata && !(adata->state&STATE_READ))
			break;
		node = articlelist_node_next(node, FALSE);
		if (!node) /* wrap */
			node = gtk_ctree_node_nth (tree, 0);
		all_read = node == start;
		if (all_read)
			break;
	}

	if (!all_read)
	{
		/* select it */
		if ( node!=NULL )
			articlelist_select_and_read_node (node);
		else {
			articlelist_unselect_all ();
			articlelist_read_next_unread_article();
		}
	}
}
void
articlelist_read_next_thread (void)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* top = gtk_ctree_node_nth (tree, 0);
	GtkCTreeNode* node = NULL;

	/* if nothing to read... */
	if ( !GTK_CLIST(tree)->rows )
		return;

	if (( node = articlelist_get_selected_node () ))
	{
		while (node!=top && GTK_CTREE_ROW(node)->parent!=top)
			node = GTK_CTREE_ROW(node)->parent;
		if (node==top)
			return;
		node = GTK_CTREE_ROW(node)->sibling;
	}
	else if (GTK_CLIST(tree)->rows > 1)
	{
		node = gtk_ctree_node_nth (tree, 1);
	}


	/* read it. */
	if (node!=NULL)
		articlelist_select_and_read_node (node);
	else {
		articlelist_unselect_all ();
		articlelist_read_next_thread ();
	}
}

void
articlelist_read_prev_article (void)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* top = gtk_ctree_node_nth (tree, 0);
	const int row_qty = GTK_CLIST(tree)->rows;
	GtkCTreeNode* node = NULL;
	gboolean wrap = FALSE;

	/* if nothing to read... */
	if ( !row_qty )
		return;

	/* get a starting node... */
	node = articlelist_get_selected_node (); 
	if ( !node || articlelist_node_prev(node,TRUE)==top ) {
		wrap = TRUE;
		node = gtk_ctree_node_nth (tree, row_qty-1 );
	}

	/* if we didn't wrap, we need to move backwards by hand */
	if ( !wrap )
		node = articlelist_node_prev (node, FALSE);

	/* read it. */
	if ( node!=NULL )
		articlelist_select_and_read_node (node);
}
void articlelist_read_prev_unread_article(void)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* top = gtk_ctree_node_nth (tree, 0);
	const int row_qty = GTK_CLIST(tree)->rows;
	GtkCTreeNode* node = NULL;
	GtkCTreeNode* start = NULL;
	gboolean all_read = FALSE;

	/* if nothing to read... */
	if (!row_qty)
		return;

	/* get starting point */	
	node = articlelist_get_selected_node ();
	if (!node || articlelist_node_prev(node,TRUE)==top) {
		node = gtk_ctree_node_nth (tree, row_qty-1); /* wrap */
	} else
		node = articlelist_node_prev(node, FALSE);

	/* move to the previous unread article (if any) */
	start = node;
	all_read = FALSE;
	for (;;) {
		const article_data* adata = (const article_data*)
			gtk_ctree_node_get_row_data (tree, node);
		if (adata && !(adata->state&STATE_READ))
			break;
		node = articlelist_node_prev(node, FALSE);
		if (!node) /* wrap */
			node = gtk_ctree_node_nth (tree, row_qty-1);
		all_read = node == start;
		if (all_read)
			break;
	}

	/* read it */
	if (!all_read && node!=NULL)
		articlelist_select_and_read_node (node);
}
void articlelist_read_prev_thread(void)
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* top = gtk_ctree_node_nth (tree, 0);
	const int row_qty = GTK_CLIST(tree)->rows;
	GtkCTreeNode* node = NULL;
	gboolean wrap = FALSE;

	if ( !row_qty )
		return;

	/* get starting point */	
	node = articlelist_get_selected_node ();
	if ( node )
		while ( node!=top && GTK_CTREE_ROW(node)->parent!=top )
			node = GTK_CTREE_ROW(node)->parent;

	/* either no node selected or we're at the first thread; wrap */
	if ( !node || articlelist_node_prev(node,TRUE)==top ) {
		wrap = TRUE;
		node = gtk_ctree_node_nth ( tree, row_qty-1 );
		while ( node!=top && GTK_CTREE_ROW(node)->parent!=top )
			node = GTK_CTREE_ROW(node)->parent;
	}

	/* if we didn't wrap around, we need to back up one. */
	if ( !wrap )
		node = articlelist_node_prev(node, TRUE);

	/* read it. */
	if ( node!=NULL )
		articlelist_select_and_read_node (node);
}


const article_data*
articlelist_get_previous_article ( void )
{
	GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	const char* message_id = NULL;

	/* get current message id */
	message_id = articlelist_get_selected_message_id ( );
	if (!message_id) message_id = articlelist_get_row_message_id (GTK_CLIST(tree)->rows-1);
	if (!message_id) return NULL;

	/* return the adata */
	return articlelist_get_relative_article (message_id, -1);
}

static void
articlelist_button_press (GtkWidget* widget, GdkEventButton* bevent)
{
	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name (GTK_OBJECT(widget),
					      "button_press_event");

	if (bevent->button == 1)
	{
		/* if in paned mode and one_click_preview set,
		 * a single click shows message 
		 */
		if (one_click_preview && viewmode==GUI_PANED )
		{
			const article_data *adata = NULL;
			int row=0, column=0;
			gtk_clist_get_selection_info (GTK_CLIST(widget), bevent->x, bevent->y, &row, &column);
			adata = (const article_data*) gtk_clist_get_row_data (GTK_CLIST(Pan.article_ctree), row);
			if (adata)
				queue_add(QUEUE_ITEM(queue_item_body_new(
					my_server, my_group, message_new(my_server, my_group, adata))));
		}
		
		/* notebook layout uses double clicking */
		if ( bevent->type == GDK_2BUTTON_PRESS )
		{
			const article_data *adata = NULL;
			int row=0, column=0;
			gtk_signal_emit_stop_by_name (GTK_OBJECT(widget), "button_press_event");
			gtk_clist_get_selection_info ( GTK_CLIST(widget), bevent->x, bevent->y, &row, &column );
			adata = (const article_data*) gtk_clist_get_row_data (GTK_CLIST(Pan.article_ctree), row);
			if (adata) {
				gui_page_set (2, text);
				queue_add(QUEUE_ITEM(queue_item_body_new(
					my_server, my_group, message_new(my_server, my_group, adata))));
				//text_set_from_message_id (my_server, my_group, message_id, NNTP_ARTICLE_READ, FALSE );
//				gtk_widget_grab_focus ( text );
			}
		}
	}
	else if ( bevent->button == 3 )
	{
		const gboolean sel = GTK_CLIST(Pan.article_ctree)->selection != NULL;
		const gboolean articles_available = GTK_CLIST(Pan.article_ctree)->rows != 0;

		GList* l = GTK_MENU_SHELL(article_ctree_menu)->children;

		pan_lock();
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* open */
		//l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* print*/
		l=l->next; /* separator */
		l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* reply to sender */
		l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* reply to group */
		l=l->next; /* separator */
		l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* save attachment */
		l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* open attachment */
		l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* delete */
		l=l->next; /* separator */
		l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* mark as read */
		l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel); /* mark as unread */
		l=l->next; /* separator */
		l=l->next; gtk_widget_set_sensitive (GTK_WIDGET(l->data), articles_available); /* mark all read */

		gtk_menu_popup (GTK_MENU (article_ctree_menu), NULL, NULL, NULL,
				NULL, bevent->button, bevent->time);
		pan_unlock();
	}
}


/**********************************************************
 * The following stuff implements the "find" functionality
 **********************************************************/

void
find_dialog_destroy_cb (void)
{
	pan_lock();
	gtk_widget_destroy(find_dialog);
	pan_unlock();

	find_dialog = NULL;
}




typedef struct
{
	gboolean unread_only;
	gboolean case_sensitive;

	const gchar *author;
	const gchar *subject;
} 
header_compare_data;

static
gboolean
compare_article (
	const article_data* adata,
	const header_compare_data* header )
{
	int flags = header->case_sensitive ? 0 : PAN_CASEFOLD;

	/* if it ain't an article, it ain't a match */
	if (!adata)
		return -1;

	/* make sure the unread state (if user specified) matches... */
	if (header->unread_only && (adata->state&STATE_READ))
		return -1;

	/* make sure the author (if user specified one) matches.. */
	if (header->author && fnmatch(header->author,adata->author,flags))
		return -1;

	/* make sure the subject (if user specified one) matches.. */
	if (header->subject && fnmatch(header->subject,adata->subject,flags))
		return -1;

	/* hooray, a match */
	return 0;
}

static
gint
find_compare_func (
	gconstpointer a,
	gconstpointer b )
{
	gint retval = -1;

	g_assert ( b!=NULL );

	if ( a!=NULL )
	{
		const article_data *adata =
			(const article_data*)a;
		const header_compare_data* header =
			(const header_compare_data*)b;
		retval = compare_article (adata, header);
	}

	return retval;
}




static
void
find_article (
	const gchar* author,
	const gchar* subject,
	gboolean case_sensitive,
	gboolean unread_only )
{
	gchar *author_tmp = author ? g_strdup_printf ("*%s*", author) : NULL;
	gchar *subject_tmp = subject ? g_strdup_printf ("*%s*", subject) : NULL;
	GtkCTreeNode* n = NULL;
	GtkCTreeNode *node = NULL;
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );
	header_compare_data header_data;

	g_return_if_fail (my_server!=NULL);
	g_return_if_fail (my_group!=NULL);

	header_data.author = author_tmp;
	header_data.subject = subject_tmp;
	header_data.unread_only = unread_only;
	header_data.case_sensitive = case_sensitive;

	node = gtk_ctree_find_by_row_data_custom (
		tree,
		find_article_next_pos,
		&header_data, 
		(GCompareFunc)find_compare_func );

	g_free (author_tmp);
	g_free (subject_tmp);

	if ( !node ) {
		find_article_next_pos = NULL;
		return;
	}

	/* select the node.. */
	articlelist_set_selected_node (node);

	/* Find next node, paying attention to recursion */
	if (GTK_CTREE_ROW(node)->children) {
		find_article_next_pos = GTK_CTREE_ROW(node)->children;
	} else {
		for (n=node; n!=NULL; n=GTK_CTREE_ROW(n)->parent) {
			if (GTK_CTREE_ROW(n)->sibling) {
				find_article_next_pos = GTK_CTREE_ROW(n)->sibling;
				break;
			}
		}
	}
}

static gchar* find_article__author = NULL;
static gchar* find_article__subject = NULL;
static gboolean find_article__case_sensitive = FALSE;
static gboolean find_article__unread_only = FALSE;

static void
find_dialog_clicked_cb (
	GnomeDialog* dialog,
	int button,
	GtkWidget* table )
{
 	g_return_if_fail ( button==0 || button==1 );
	g_return_if_fail ( table!=NULL );

	if ( button==1 ) /* close button */
	{
		pan_lock();
		gtk_widget_unmap ( GTK_WIDGET(find_dialog) );
		pan_unlock();
	}
	else if ( button==0 ) /* "find next" button */
	{
		GtkWidget* author_entry;
		GtkWidget* subject_entry;
		GtkWidget* case_sensitive_checkbox;
		GtkWidget* unread_only_checkbox;
		gchar* author = 0;
		gchar* subject = 0;
		gboolean case_sensitive;
		gboolean unread_only;

		/* get author to search for, or NULL if none... */
		author_entry = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "author" );
		g_return_if_fail ( author_entry!=NULL );
		author = g_strdup ( gtk_entry_get_text ( GTK_ENTRY(author_entry) ) );
		g_strstrip ( author );
		if ( !*author ) {
			g_free ( author );
			author = NULL;
		}

		/* get subject to search for, or NULL if none... */
		subject_entry = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "subject" );
		g_return_if_fail ( subject_entry!=NULL );
		subject = g_strdup ( gtk_entry_get_text ( GTK_ENTRY(subject_entry) ) );
		g_strstrip ( subject );
		if ( !*subject ) {
			g_free ( subject );
			subject = NULL;
		}

		/* get case sensitivity... */
		case_sensitive_checkbox = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "case" );
		g_return_if_fail ( case_sensitive_checkbox!=NULL );
		case_sensitive = gtk_toggle_button_get_active ( GTK_TOGGLE_BUTTON(case_sensitive_checkbox) );

		/* get case sensitivity... */
		unread_only_checkbox = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "unread" );
		g_return_if_fail ( unread_only_checkbox!=NULL );
		unread_only = gtk_toggle_button_get_active ( GTK_TOGGLE_BUTTON(unread_only_checkbox) );

		/* update the find_article fields with these widget values */
		g_free ( find_article__author );
		find_article__author = author;
		g_free ( find_article__subject );
		find_article__subject = subject;
		find_article__case_sensitive = case_sensitive;
		find_article__unread_only = unread_only;

		articlelist_find_next_cb ( );
	}
}

void
articlelist_find_next_cb (void)
{
	/* user may have hit "find again" without hitting "find" first */
	if (!find_article__subject && !find_article__author) {
		articlelist_find_text_cb();
		return;
	}

	/* got a subject and/or an author.. let's go find 'em */
	find_article (
		find_article__author,
		find_article__subject,
		find_article__case_sensitive,
		find_article__unread_only );
}


void
articlelist_find_text_cb (void)
{
	GtkWidget *table;
	GtkWidget *author_label, *author_entry;
	GtkWidget *subject_label, *subject_entry;
	GtkWidget *case_sensitive_checkbutton;
	GtkWidget *unread_only_checkbutton;

	/* There can be only one! (find window) */
	if (find_dialog) {
		if ( !GTK_WIDGET_MAPPED(GTK_WIDGET(find_dialog)) )
			gtk_widget_map ( GTK_WIDGET(find_dialog) );
		gdk_window_raise (GTK_WIDGET(find_dialog)->window);
		return;
	}
	
	find_dialog = gnome_dialog_new ( _("Find Message"),
					 GNOME_STOCK_PIXMAP_FORWARD,
					 GNOME_STOCK_PIXMAP_CLOSE, NULL);
	gnome_dialog_set_default ( GNOME_DIALOG(find_dialog), 0 );

	table = gtk_table_new ( 4, 2, FALSE );
//	gtk_container_set_border_width ( GTK_CONTAINER(table), GNOME_PAD_SMALL );
//        gtk_table_set_row_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
//        gtk_table_set_col_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);

	subject_label = gtk_label_new ( _("Subject:") );
	gtk_table_attach ( GTK_TABLE(table),
			   subject_label,
			   0, 1,
			   0, 1,
			   0,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	subject_entry = gtk_entry_new ( );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "subject", subject_entry );
	gtk_table_attach ( GTK_TABLE(table),
			   subject_entry,
			   1, 2,
			   0, 1,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	author_label = gtk_label_new ( _("Author:") );
	gtk_table_attach ( GTK_TABLE(table),
			   author_label,
			   0, 1,
			   1, 2,
			   0,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	author_entry = gtk_entry_new ( );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "author", author_entry );
	gtk_table_attach ( GTK_TABLE(table),
			   author_entry,
			   1, 2,
			   1, 2,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	unread_only_checkbutton = gtk_check_button_new_with_label ( "Unread Only" );
	gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON(unread_only_checkbutton), FALSE );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "unread", unread_only_checkbutton );
	gtk_table_attach ( GTK_TABLE(table),
			   unread_only_checkbutton,
			   0, 2,
			   2, 3,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   GTK_EXPAND | GTK_FILL,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	case_sensitive_checkbutton = gtk_check_button_new_with_label ( "Case Sensitive" );
	gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON(case_sensitive_checkbutton), FALSE );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "case", case_sensitive_checkbutton );
	gtk_table_attach ( GTK_TABLE(table),
			   case_sensitive_checkbutton,
			   0, 2,
			   3, 4,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   GTK_EXPAND | GTK_FILL,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	
//	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(find_dialog)->vbox), table );
	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (find_dialog)->vbox), table, FALSE, FALSE, 0);
	
	gtk_signal_connect (GTK_OBJECT(find_dialog), "clicked",
			    GTK_SIGNAL_FUNC(find_dialog_clicked_cb),
			    table);
	gtk_signal_connect (GTK_OBJECT(find_dialog), "destroy",
			    GTK_SIGNAL_FUNC(find_dialog_destroy_cb), NULL);

	gtk_widget_grab_focus (subject_entry);
	gui_popup_draw (find_dialog, Pan.window);
}


/*--------------------------------------------------------------------
 * clear the ctree of all messages
 *--------------------------------------------------------------------*/

static void
clear_article_buffer (void)
{
	if (article_header_qty)
	{
		int i;
		for (i=0; i<article_header_qty; ++i)
			article_free (article_headers[i]);
		g_free (article_headers);
		article_header_qty = 0;
	}
}

void
articlelist_clear (void)
{
	GtkCList *list = GTK_CLIST(Pan.article_ctree);
	const int row_qty = list->rows;

	if (row_qty)
	{
		pan_lock();
		gtk_clist_freeze (list);
		gtk_clist_clear (list);
		gtk_clist_thaw (list);
		pan_unlock();
	}

	if (active_group_header_db != NULL)
		ahdb_unref (active_group_header_db);
	my_group = NULL;
	my_server = NULL;

	gui_set_title (NULL);
	clear_article_buffer();
}

/*--------------------------------------------------------------------
 * load article listing
 *--------------------------------------------------------------------*/
void
articlelist_load (
	server_data *sdata,
	group_data *gdata,
	StatusItem *item)
{
	article_db db = (article_db)0;
	GSList *alist = NULL;
	int total = 0;
	int loaded_total = 0;

	/* argument sanity check */
	g_return_if_fail (gdata != NULL);

	/* step 1: load article headers */

	total = group_get_attrib_i (sdata, gdata, "Total");
	status_item_emit_status (item, _("Loading article headers"));
	status_item_emit_init_steps (item, total*4);
	db = ahdb_ref (sdata, gdata);
	alist = ahdb_get_all (db, status_item_next_step_gfunc, item);
	ahdb_unref (db);

	/* check totals */
	loaded_total = g_slist_length (alist);
	if (total != loaded_total) {
		g_warning (
			_("Pan thought there were %d articles; instead we got %d"),
			total,
			loaded_total );
		group_set_attrib_i (sdata, gdata, "Total", total=loaded_total);
	}

	if (!alist)
	{
		choose_download_dialog (sdata, gdata);
	} 
	else
	{
		log_add_va (
			_("Loaded %d article headers for group %s"),
			total,
			gdata->name );
		articlelist_set_contents (sdata, gdata, alist, total, item);
	}

	status_item_emit_progress (item, 0);
}

static GSList*
killfile_articles (
	const group_data* gdata,
	GSList* alist,
	int* alist_total,
	StatusItem *item)
{
	GSList* prev = NULL;
	GSList* cur = alist;

	while ( cur != NULL )
	{
		article_data* adata = (article_data*) cur->data;

		const MessageFilterAction action =
			message_filter_get_action ( gdata, adata );

		if ( action & MESSAGE_FILTER_ACTION_DELETE )
		{
			/* mark for deletion */
			GSList* del1me = cur;

			/* unlink */
			if ( prev )
				prev->next = cur->next;
			else
				alist = cur->next;
			cur = cur->next;

			/* delete article from list */
			--*alist_total;
			g_slist_free_1 ( del1me );
			article_free ( adata );
			del1me->data = NULL;
		}
		else
		{
			if ( action & MESSAGE_FILTER_ACTION_COLOR )
			{
				adata->state |= STATE_KILLFILE;
			}

			prev = cur;
			cur = cur->next;
		}

		status_item_emit_next_step (item);
	}

	return alist;
}

void
articlelist_queue_reload (void)
{
	if (my_group != NULL)
	{
		queue_add (QUEUE_ITEM(queue_item_headers_new(
			(server_data*)my_server, TRUE,
			(group_data*)my_group,
			HEADERS_LOCAL)));
	}
}

static void
update_sort_menu (int sort_mode)
{
	GtkWidget *menu = NULL;
	int menu_item_index = GPOINTER_TO_INT(g_hash_table_lookup (
		sort_type_to_menu_index, GINT_TO_POINTER(sort_mode)));

	/* why the fsck doesn't gtk have a way to select an option menu item
	   without the activation event?  garrggh. */
	pan_lock();
	menu = gtk_option_menu_get_menu (GTK_OPTION_MENU(sort_option_menu));
	gtk_object_ref(GTK_OBJECT(menu));
	gtk_option_menu_remove_menu (GTK_OPTION_MENU(sort_option_menu));
	gtk_menu_set_active (GTK_MENU(menu), menu_item_index);
	gtk_option_menu_set_menu (GTK_OPTION_MENU(sort_option_menu), menu);
	gtk_object_unref(GTK_OBJECT(menu));
	pan_unlock();
}

void
articlelist_set_contents (
	server_data* sdata,
	group_data* gdata,
	GSList* alist,
	int alist_total,
	StatusItem *item)
{
	/* get sorting information */
	sort_type = group_get_attrib_i (sdata, gdata,"sort_style");
	if (!sort_type || abs(sort_type)>ARTICLE_SORT_TYPES_QTY)
		sort_type = -ARTICLE_SORT_DATE;

	update_sort_menu (sort_type);

	/* update local state */
	my_server = sdata;
	my_group = gdata;
	if (active_group_header_db != NULL)
		ahdb_unref (active_group_header_db);
	if (my_group != NULL)
		active_group_header_db = ahdb_ref (sdata, gdata);

	/* killfile */
	status_item_emit_status (item, _("Applying Killfile"));
	alist = killfile_articles (gdata, alist, &alist_total, item);

	/* thread */
	status_item_emit_status (item, _("Threading article list"));
	clear_article_buffer ();
	thread_articles (
		alist, alist_total,
		&article_headers, &article_header_qty, item);
	alist = NULL;

	/* sort */
	sort_articles (
		article_headers, article_header_qty,
		abs(sort_type), sort_type>0);

	/* add to tree */
	status_item_emit_status (item, _("Displaying article list"));
	articlelist_update (gdata,
		(article_data**)article_headers, article_header_qty, item);
}

const server_data*
articlelist_get_current_server (void)
{
	return my_server;
}
group_data*
articlelist_get_current_group (void)
{
	return my_group;
}


//============================================================================

typedef struct
{
	GtkCTreeNode* node;
	article_db db;
	int total_qty;
	int read_qty;
	GFunc func;
}
CountStruct;

static gboolean
ancestor_is_selected ( const GtkCTree* tree, const GtkCTreeNode* node )
{
	const GList* list =  GTK_CLIST(tree)->selection;
	if ( !list )
		return FALSE;

	for ( ;; ) {
		node = GTK_CTREE_ROW(node)->parent;
		if ( !node )
			return FALSE;
		if ( g_list_index ( (GList*)list, (gpointer)node ) != -1 )
			return TRUE;
	}

        pan_warn_if_reached  ( );
}

static void
articlelist_db_erase_article_gfunc (gpointer data, gpointer user_data)
{
	const article_data* adata = (article_data*)data;
	CountStruct* count = (CountStruct*)user_data;
	ahdb_erase (count->db, adata, NULL);

	if (count->node == find_article_next_pos)
		find_article_next_pos = NULL;
}


static void
articlelist_db_mark_unread_gfunc (gpointer data, gpointer user_data)
{
	article_data* adata = (article_data*)data;
	if (adata->state & STATE_READ)
	{
		CountStruct* count = (CountStruct*)user_data;
		adata->state &= ~STATE_READ;
		ahdb_save (count->db, adata, FALSE);
		articlelist_update_node_fast (count->node, adata);
	}
}


static void
articlelist_db_mark_read_gfunc (gpointer data, gpointer user_data)
{
	article_data* adata = (article_data*)data;
	if (!( adata->state & STATE_READ ))
	{
		CountStruct* count = (CountStruct*)user_data;
		adata->state |= STATE_READ;
		ahdb_save (count->db, adata, FALSE);
		articlelist_update_node_fast (count->node, adata);
	}
}


static void
count_articles ( GtkCTree* ctree,
		 GtkCTreeNode* node,
		 gpointer data )
{
	CountStruct* count = (CountStruct*)data;
	article_data* adata = NULL;

	/* sanity checking */
	g_return_if_fail ( count!=NULL );
	g_return_if_fail ( ctree!=NULL );
	g_return_if_fail ( node!=NULL );

	/* get the article */
	adata = (article_data*) gtk_ctree_node_get_row_data (ctree, node);
	if (!adata) {
		if (node == gtk_ctree_node_nth(ctree,0)) /* name-of-group */
			return;
		else g_assert (adata!=NULL);
	}

	/* tally up the read/total */
	++count->total_qty;
	if (adata->state & STATE_READ)
		++count->read_qty;

	/* call the client-specified function */
	count->node = node;
	(*count->func)(adata, data);
}

void
articlelist_selected_delete ( void )
{
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );
	const server_data* sdata = my_server;
	const group_data* group = my_group;
	GSList* l = NULL;
	GSList* prune_branches = NULL;
	CountStruct count;

	count.read_qty = 0;
	count.total_qty = 0;
	count.db = ahdb_ref ( sdata, group );
	count.func = articlelist_db_erase_article_gfunc;

	gtk_clist_freeze ( GTK_CLIST(tree) );

	/* first we update the article header db... */
	for ( l=(GSList*)GTK_CLIST(tree)->selection; l!=NULL; l=l->next ) {
		if ( !ancestor_is_selected ( tree, GTK_CTREE_NODE(l->data) ) ) {
			gtk_ctree_post_recursive ( tree, GTK_CTREE_NODE(l->data), count_articles, &count );
			prune_branches = g_slist_prepend ( prune_branches, l->data );
		}
	}

	/* then we update the group attributes db... */
	group_add_attrib_i (sdata,  group, "Read", -count.read_qty );
	group_add_attrib_i (sdata,  group, "Total", -count.total_qty );

	/* then finally we update the ui. */
	gtk_ctree_unselect_recursive ( tree, NULL );
	for ( l=prune_branches; l!=NULL; l=l->next )
		gtk_ctree_remove_node ( tree, GTK_CTREE_NODE(l->data) );

	gtk_clist_thaw ( GTK_CLIST(tree) );

	/* cleanup */
	g_slist_free ( prune_branches );
	ahdb_sync ( count.db );
	ahdb_unref ( count.db );
}

static void
articlelist_mark_read (GSList* l)
{
	GtkCTree* tree = GTK_CTREE (Pan.article_ctree);
	CountStruct count;
	int newly_read;

	/* argument sanity checking */
	g_return_if_fail (l!=NULL);

	/* setup */
	count.read_qty = 0;
	count.total_qty = 0;
	count.db = ahdb_ref (my_server, my_group);
	count.func = articlelist_db_mark_read_gfunc;

	/* mark them all as read */
	gtk_clist_freeze ( GTK_CLIST(tree) );
	for ( ; l!=NULL; l=l->next )
		if ( !ancestor_is_selected ( tree, GTK_CTREE_NODE(l->data) ) )
			gtk_ctree_post_recursive ( tree, GTK_CTREE_NODE(l->data), count_articles, &count );
	gtk_ctree_unselect_recursive ( tree, NULL );
	gtk_clist_thaw ( GTK_CLIST(tree) );

	/* update the group attributes */
	newly_read = count.total_qty - count.read_qty;
	group_add_attrib_i (my_server, my_group, "Read", newly_read);

	/* cleanup */
	ahdb_sync ( count.db );
	ahdb_unref ( count.db );
}


void
articlelist_selected_mark_read ( void )
{
	articlelist_mark_read ( (GSList*) GTK_CLIST(Pan.article_ctree)->selection );
}

void
articlelist_all_mark_read ( void )
{
	GSList* l = g_slist_prepend ( NULL, gtk_ctree_node_nth ( GTK_CTREE(Pan.article_ctree), 0 ) );
	articlelist_mark_read ( l );
	g_slist_free ( l );
}

void
articlelist_selected_mark_unread ( )
{
	GtkCTree* tree = GTK_CTREE ( Pan.article_ctree );
	const server_data* sdata = my_server;
	const group_data* group = my_group;
	GSList* l;
	CountStruct count;

	/* setup */
	count.read_qty = 0;
	count.total_qty = 0;
	count.db = ahdb_ref ( sdata, group );
	count.func = articlelist_db_mark_unread_gfunc;

	/* mark them all as read */
	gtk_clist_freeze ( GTK_CLIST(tree) );
	for ( l=(GSList*)GTK_CLIST(tree)->selection; l!=NULL; l=l->next )
		if ( !ancestor_is_selected ( tree, GTK_CTREE_NODE(l->data) ) )
			gtk_ctree_post_recursive ( tree, GTK_CTREE_NODE(l->data), count_articles, &count );
	gtk_ctree_unselect_recursive ( tree, NULL );
	gtk_clist_thaw ( GTK_CLIST(tree) );

	/* update the group attributes */
	group_add_attrib_i (sdata,  group, "Read", -count.read_qty );

	/* cleanup */
	ahdb_unref ( count.db );
}

static void
articlelist_selected_decode_impl (gboolean open)
{
	GtkCTree* tree = GTK_CTREE (Pan.article_ctree);
	GSList* l = NULL;

	//g_return_if_fail (NULL); /* I'll fix this tomorrow */

	/* sanity checks... */
	g_return_if_fail (my_server!=NULL);
	g_return_if_fail (my_group!=NULL);

	/* loop through all the selected articles */
	gtk_clist_freeze (GTK_CLIST(tree));
	for (l=(GSList*)GTK_CLIST(tree)->selection; l!=NULL; l=l->next)
	{
		/* build an array of this node and all its children. */
		GSList *tmp = NULL;
		GSList *list = NULL;
		GtkCTreeNode* node = GTK_CTREE_NODE(l->data);
		article_data* adata = (article_data*) gtk_ctree_node_get_row_data (tree,node);
		if (adata == NULL)
			continue;
		list = g_slist_append (list, adata->message_id);
		for (tmp=adata->threads; tmp!=NULL; tmp=tmp->next)
			list = g_slist_append (list, ((article_data*)tmp->data)->message_id);

		/* queue for decode */
		queue_add (QUEUE_ITEM(
			queue_item_decode_new(
				(server_data*)my_server,
				(group_data*)my_group,
				adata->subject, list, open)));

		/* update the list to show this is queued */
		adata->state |= STATE_DECODE_QUEUED;
		articlelist_update_node_fast (GTK_CTREE_NODE(l->data), adata);
	
		/* cleanup this loop */
		g_slist_free (list);
	}

	/* cleanup */
	gtk_ctree_unselect_recursive ( tree, NULL );
	gtk_clist_thaw ( GTK_CLIST(tree) );
}
	
void
articlelist_selected_decode ( void )
{		
	articlelist_selected_decode_impl (FALSE);
}

void
articlelist_selected_open ( void )
{
	articlelist_selected_decode_impl (TRUE);
}

void
articlelist_select_all ( void )
{
	pan_lock ();
	gtk_ctree_select_recursive (GTK_CTREE(Pan.article_ctree), NULL);
	pan_unlock ();
}


//============================================================================


static
GnomePixmap*
get_article_pixmap (
	const article_data* adata )
{
	g_return_val_if_fail ( adata!=NULL, NULL );

	if (adata->state & STATE_ERROR)
		return error_pixmap;

	if (adata->state & STATE_DECODED)
		return binary_decoded_pixmap;

#if 0
	if ((adata->state & STATE_UNREAD_CHILDREN)
	    && (!(adata->state & STATE_MULTIPART_ALL))
	    && (!(adata->state & STATE_MULTIPART_SOME)))
		return unread_children_pixmap;
#endif

	if (adata->state & STATE_READ)
		return article_read_pixmap;

	if (adata->state & STATE_DOWNLOAD_QUEUED)
		return download_queue_pixmap;

	if (adata->state & STATE_DECODE_QUEUED)
		return decode_queue_pixmap;

	if (adata->state & STATE_MULTIPART_ALL)
		return mp_complete_pixmap;

	if (adata->state & STATE_MULTIPART_SOME)
		return mp_incomplete_pixmap;

	return NULL;
}


static void
articlelist_set_node_style (
	GtkCTreeNode* node,
	const article_data* adata)
{
	GtkCTree* tree = GTK_CTREE (Pan.article_ctree);

	pan_lock();

	if (!unread_style)
	{
		GdkFont* font = NULL;
		if (articlelist_font != NULL)
			font = gdk_fontset_load (articlelist_font);

		killfile_style = gtk_style_copy (
				gtk_widget_get_style (GTK_WIDGET (Pan.window)));
		killfile_style->fg[0] = killfile_color;

		unread_style = gtk_style_copy (
				gtk_widget_get_style (GTK_WIDGET (Pan.window)));
		unread_style->fg[0] = unread_color;

		read_style = gtk_style_copy (
				gtk_widget_get_style (GTK_WIDGET (Pan.window)));
		read_style->fg[0] = read_color;

		new_style = gtk_style_copy (
				gtk_widget_get_style (GTK_WIDGET (Pan.window)));
		new_style->fg[0] = new_color;

		if (font) {
			killfile_style->font = font;
			unread_style->font = font;
			read_style->font = font;
			new_style->font = font;
		}
	}

	if (articlelist_font)
	{
		if (adata->state & STATE_KILLFILE)
			gtk_ctree_node_set_row_style (tree, node, killfile_style);
		else if ((adata->state & STATE_READ) || (adata->state & STATE_DECODED))
			gtk_ctree_node_set_row_style (tree, node, read_style);
		else if (adata->state & STATE_NEW)
			gtk_ctree_node_set_row_style (tree, node, new_style);
		else
			gtk_ctree_node_set_row_style (tree, node, unread_style);
	}	
	else /* If the font isn't specified by the user, just set the color */
	{
		if (adata->state & STATE_KILLFILE)
			gtk_ctree_node_set_foreground (tree, node, &killfile_color);
		else if ((adata->state & STATE_READ) || (adata->state & STATE_DECODED))
			gtk_ctree_node_set_foreground (tree, node, &read_color);
		else if (adata->state & STATE_NEW)
			gtk_ctree_node_set_foreground (tree, node, &new_color);
		else
			gtk_ctree_node_set_foreground (tree, node, &unread_color);
	}

	pan_unlock();
}

static GtkCTreeNode*
add_article_to_ctree (
	GtkCTree* tree,
	GtkCTreeNode* parent,
	article_data* adata,
	StatusItem *item,
	gboolean expanded)
{
	char *text[4];
	char line_count_buf[16];
	GtkCTreeNode *node;
	GnomePixmap* icon = get_article_pixmap (adata);
        char timebuf[32] = { '\0' };
        const struct tm local_tm = *localtime (&adata->date);
	gulong lcountsum = 0;

        /* this is sortable with a simple strcmp */
        strftime (timebuf, sizeof(timebuf), "%Y/%m/%d %H:%M", &local_tm);

	/* line count */
	lcountsum = adata->linecount;
	if (adata->state & STATE_MULTIPART_ALL) {
		GSList *l;
		for (l=adata->threads; l; l=l->next)
			lcountsum += ((article_data *)l->data)->linecount;
	}
	commatize_ulong ( lcountsum, line_count_buf );

	/* set text */
	text[0] = adata->subject;
	text[1] = line_count_buf;
	text[2] = timebuf;
	text[3] = adata->author;

	status_item_emit_next_step (item);

	pan_lock();
	node = gtk_ctree_insert_node (
		tree, parent, NULL,
		text, 5,
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		FALSE, expanded);
	gtk_ctree_node_set_row_data (tree, node, adata);
	pan_unlock();
	articlelist_set_node_style (node, adata);

	return node;
}

static void
build_article_ctree_recursive (
	GtkCTree* tree,
	GtkCTreeNode* parent,
	GSList *list,
	StatusItem *item,
	gboolean expanded)
{
	GSList *pos = NULL;

	/* sanity check */
	g_return_if_fail (parent!=NULL);

        for (pos=list; pos!=NULL; pos=pos->next)
        {
		article_data* adata = (article_data*) pos->data;
		GtkCTreeNode* node = add_article_to_ctree (tree, parent, adata, item, expanded);
		if (adata->threads!=NULL)
			build_article_ctree_recursive (tree, node, adata->threads, item, TRUE);
        }
}



typedef struct
{
	const group_data *gdata;
	article_data **article_buf;
	int article_qty;
	StatusItem *item;
	int index;
	GtkCTreeNode *root;
}
BuildArticleListStruct;

static void
add_nodes (gpointer data)
{
	BuildArticleListStruct *build = (BuildArticleListStruct*) data;
	GtkCTree* tree = GTK_CTREE (Pan.article_ctree);

	while (build->index < build->article_qty)
	{
		article_data* adata = build->article_buf[build->index++];
		const gboolean has_parent = adata->parent != NULL;
		const gboolean filter_match =
			subject_filter_str==NULL ||
			!fnmatch(subject_filter_str,adata->subject,PAN_CASEFOLD);
		GtkCTreeNode* node = NULL;

		if ((!is_threaded && filter_match && adata->part<2)
		    || (is_threaded && !has_parent))
		{
			node = add_article_to_ctree (
				tree, build->root, adata, build->item, FALSE);
		}
		if (is_threaded && node!=NULL && adata->threads!=NULL
		    && !((adata->state&STATE_MULTIPART_ALL) && hide_mpart_child_nodes))
		{
			build_article_ctree_recursive (
				tree, node, adata->threads, build->item, TRUE);
		}
	}

	pan_lock();
	gtk_ctree_expand (tree, build->root);
	gtk_clist_thaw (GTK_CLIST(tree));
	gtk_widget_queue_resize (Pan.article_ctree);
	pan_unlock();
}

/* fires next_step article_qty times */
static void
articlelist_update (
	const group_data* gdata,
	article_data** article_buf,
	int article_qty,
	StatusItem *item)
{
	GtkCTree* tree = GTK_CTREE (Pan.article_ctree);
	char *text[4];

	BuildArticleListStruct build;
	build.gdata = gdata;
	build.article_buf = article_buf;
	build.article_qty = article_qty;
	build.item = item;
	build.index = 0;
	build.root = NULL;

	status_item_emit_status (item, _("Updating article display"));

	pan_lock();

	/* clear out old */
	gtk_clist_freeze (GTK_CLIST(tree));
	gtk_clist_clear (GTK_CLIST(tree));
	find_article_next_pos = NULL;

	/* add root node */
	text[0] = gdata->name;
	text[1] = text[2] = text[3] = NULL;
	build.root = gtk_ctree_insert_node (
		tree, NULL, NULL,
		text, 5, articlelist_closed_pixmap->pixmap,
		articlelist_closed_pixmap->mask,
		articlelist_open_pixmap->pixmap,
		articlelist_open_pixmap->mask, FALSE, FALSE );
	gtk_ctree_node_set_row_data (tree, build.root, NULL);
	pan_unlock();

	add_nodes (&build);
}

/***
****
****
***/

static void 
articlelist_update_node_fast (GtkCTreeNode* node, const article_data* adata)
{
	GnomePixmap* icon = get_article_pixmap (adata);

	pan_lock();
	gtk_ctree_set_node_info (
		GTK_CTREE(Pan.article_ctree), node,
		adata->subject, 5,
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		GTK_CTREE_ROW(node)->is_leaf,
		GTK_CTREE_ROW(node)->expanded );
	pan_unlock();
	articlelist_set_node_style (node, adata);
}


static GtkCTreeNode*
articlelist_get_node_from_message_id (const char* message_id, int* index)
{
        GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
	GtkCTreeNode* node = gtk_ctree_node_nth(tree, 0);
	const article_data *adata = NULL;
	int i=0;

	g_assert ( message_id != NULL );

	for ( ;; )
	{
		++i;
		node = articlelist_node_next (node, FALSE);
		if ( !node )
			return NULL;


		adata = (const article_data*) gtk_ctree_node_get_row_data(tree, node);
		if (adata && !strcmp( message_id, adata->message_id)) {
			if (index)
				*index = i;
			return node;
		}
	}

	pan_warn_if_reached  ( );
}

void
articlelist_update_node (const article_data* adata)
{
	GtkCTreeNode* node = articlelist_get_node_from_message_id (adata->message_id, NULL);
        if (node != NULL)
	{
		/* update the adata state */
		GtkCTree* tree = GTK_CTREE(Pan.article_ctree);
		article_data* real_adata =
			(article_data*) gtk_ctree_node_get_row_data (tree, node);
		real_adata->state = adata->state;

		/* repaint */
                articlelist_update_node_fast (node, real_adata);
	}
}


static void
articlelist_click_column (GtkCList* clist, int n)
{
	int type;
	switch (n) {
		case 0: type = ARTICLE_SORT_SUBJECT; break;
		case 1: type = ARTICLE_SORT_LINES; break;
		case 2: type = ARTICLE_SORT_DATE; break;
		case 3: type = ARTICLE_SORT_AUTHOR; break;
		default: pan_warn_if_reached(); type = -ARTICLE_SORT_DATE; break;
	}
	if (type == sort_type)
		type = -type;

	update_sort_menu (type);
	sort_selected_cb (NULL, GINT_TO_POINTER(type));
}

static void
article_ctree_destroy_cb (void)
{
	gtk_widget_destroy (GTK_WIDGET(articlelist_closed_pixmap));
	gtk_widget_destroy (GTK_WIDGET(articlelist_open_pixmap));
	gtk_widget_destroy (GTK_WIDGET(article_read_pixmap));
	gtk_widget_destroy (GTK_WIDGET(decode_queue_pixmap));
	gtk_widget_destroy (GTK_WIDGET(download_queue_pixmap));
	gtk_widget_destroy (GTK_WIDGET(error_pixmap));
	gtk_widget_destroy (GTK_WIDGET(binary_decoded_pixmap));
	gtk_widget_destroy (GTK_WIDGET(unread_children_pixmap));
	gtk_widget_destroy (GTK_WIDGET(mp_complete_pixmap));
	gtk_widget_destroy (GTK_WIDGET(mp_incomplete_pixmap));
}

static gchar*
sort_selected_describe (const StatusItem* item)
{
	return g_strdup ("Sorting Articles");
}

/*ccc*/
static void
sort_selected_thread (void* data)
{
	StatusItem *item = NULL;

	if ( !my_group ) /* articlelist is idle */
		return;

	/* create a new status item to get sort/thread messages */
	item = STATUS_ITEM(status_item_new(sort_selected_describe));
	pan_object_ref(PAN_OBJECT(item));
	gui_add_status_item (item);

	/* sort the articles & update the list */
	status_item_emit_init_steps (item, article_header_qty);
       	sort_type = GPOINTER_TO_INT(data);
	group_set_attrib_i (my_server, my_group, "sort_style", sort_type);
	sort_articles (
		article_headers, article_header_qty, abs(sort_type), (sort_type>0));
	articlelist_update (
		my_group, (article_data**)article_headers, article_header_qty, item);

	/* clean out the status item */
	gui_remove_status_item (item);
	pan_object_unref(PAN_OBJECT(item));
}

static void
sort_selected_cb (GtkWidget* w, gpointer data)
{
	pthread_t thread;
	pthread_create (&thread, NULL, (void*)sort_selected_thread, data);
	pthread_detach (thread);
}

static void
article_filter_cb (GtkWidget* w, gpointer data)
{
	if (subject_filter_str)
		g_free (subject_filter_str);
	subject_filter_str =
		g_strdup_printf("*%s*",gtk_entry_get_text(GTK_ENTRY(w)));
	sort_selected_cb (NULL, GINT_TO_POINTER(ARTICLE_SORT_SUBJECT));
}

static void
threaded_toggled_cb (GtkToggleButton *tb, gpointer user_data)
{
	is_threaded = gtk_toggle_button_get_active (tb);
	sort_selected_cb (NULL, GINT_TO_POINTER(sort_type));
}

GtkWidget*
articlelist_create_toolbar (void)
{
	GtkWidget *w = NULL;
	GtkWidget *w2 = NULL;
	GtkWidget *item = NULL;
	GtkWidget *toolbar = NULL;
	GtkWidget *menu = NULL;
	int i;
	int sel_index;

	const gchar* names[] = {
		N_("+Author"),
		N_("-Author"),
		N_("+Date"),
		N_("-Date"),
		N_("+Subject"),
		N_("-Subject"),
		N_("+Line Count"),
		N_("-Line Count")
	};
	const gint sort[] = {
		 ARTICLE_SORT_AUTHOR,
		-ARTICLE_SORT_AUTHOR,
		 ARTICLE_SORT_DATE,
		-ARTICLE_SORT_DATE,
		 ARTICLE_SORT_SUBJECT,
		-ARTICLE_SORT_SUBJECT,
		 ARTICLE_SORT_LINES,
		-ARTICLE_SORT_LINES
	};

	const int row_qty = sizeof(sort) / sizeof(sort[0]);

	/* create the hash table... */
	sort_type_to_menu_index = g_hash_table_new (
		g_direct_hash, g_direct_equal);

	/* create the toolbar... */
	toolbar = gtk_toolbar_new (
		GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_TEXT);

	/* "Articles" label... */
	w = gtk_label_new (_("Articles"));
	gtk_toolbar_append_widget (GTK_TOOLBAR(toolbar), w, "", "");
	gtk_toolbar_append_space (GTK_TOOLBAR(toolbar));

	/* Thread? */
	w = gtk_toggle_button_new ();
	w2 = gnome_pixmap_new_from_xpm_d (thread_xpm);
	gtk_container_add (GTK_CONTAINER(w), w2);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(w),  TRUE);
	gtk_signal_connect (GTK_OBJECT(w), "toggled", threaded_toggled_cb, NULL);
	gtk_toolbar_append_widget (GTK_TOOLBAR(toolbar), w, "Thread Articles", "");

	/* sorting optionmenu */
	w = sort_option_menu = gtk_option_menu_new ();
	gtk_widget_set (w, "width", 120, "height", 25, NULL); /* FIXME */
	menu = gtk_menu_new();
	sel_index = -1;
	for (i=0; i<row_qty; ++i) {
		item = gtk_menu_item_new_with_label (_(names[i]));
		gtk_signal_connect (GTK_OBJECT(item), "activate",
			(GtkSignalFunc)sort_selected_cb,
			GINT_TO_POINTER(sort[i]));
		gtk_menu_append (GTK_MENU(menu), item);
		g_hash_table_insert (
			sort_type_to_menu_index,
			GINT_TO_POINTER(sort[i]),
			GINT_TO_POINTER(i));
	}
	gtk_option_menu_set_menu (GTK_OPTION_MENU(w), menu);
	gtk_toolbar_append_widget (GTK_TOOLBAR(toolbar), w, "Article Sorting", "");
	w = gtk_entry_new();
	gtk_signal_connect (GTK_OBJECT (w), "activate",
			    GTK_SIGNAL_FUNC(article_filter_cb), w);
	gtk_toolbar_set_space_size (GTK_TOOLBAR(toolbar), 100);
	gtk_toolbar_append_widget (GTK_TOOLBAR(toolbar), w,
				   _("Type in a subject search string and press ENTER.  "
				     "Wildcards are allowed."), "");

	/* cleanup */
	gtk_widget_show_all (toolbar);
	gnome_app_add_toolbar (
		GNOME_APP(Pan.window),
		GTK_TOOLBAR(toolbar), "article_toolbar", 0, GNOME_DOCK_TOP, 1, 1, 1);

	return 0;
}

/*--------------------------------------------------------------------
 * generate the listing of articles, for the "Articles" tab
 *--------------------------------------------------------------------*/
GtkWidget*
create_articlelist_ctree (void)
{
	GtkWidget *s_window;
	char* titles[4];

	/* set the titles one-by-one so gettext() can work them */
	titles[0] = _("Subject");
	titles[1] = _("Lines");
	titles[2] = _("Date");
	titles[3] = _("Author");

	/* load the pixmaps */
	articlelist_closed_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(book_closed_xpm);
	articlelist_open_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(book_open_xpm);
	article_read_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(mini_page_xpm);
	decode_queue_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(bluecheck_xpm);
	download_queue_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(greencheck_xpm);
	error_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(x_xpm);
	binary_decoded_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(binary_xpm);
	unread_children_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(unread_children_xpm);
	mp_complete_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(binary_complete_xpm);
	mp_incomplete_pixmap = (GnomePixmap*)
		gnome_pixmap_new_from_xpm_d(binary_incomplete_xpm);

	cmap = gdk_colormap_get_system ();
	if (!gdk_color_alloc (cmap, &unread_color))
		g_error ("couldn't allocate color");
	if (!gdk_color_alloc (cmap, &new_color))
		g_error ("couldn't allocate color");
	if (!gdk_color_alloc (cmap, &read_color))
		g_error ("couldn't allocate color");

	Pan.article_ctree = gtk_ctree_new_with_titles (4, 0, titles);
	gtk_ctree_set_line_style (GTK_CTREE(Pan.article_ctree),
				  GTK_CTREE_LINES_DOTTED);
	gtk_widget_show (Pan.article_ctree);

	s_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(s_window),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(s_window), Pan.article_ctree);

	gtk_clist_set_selection_mode (GTK_CLIST(Pan.article_ctree),
				      GTK_SELECTION_EXTENDED);

	gtk_clist_set_column_width (GTK_CLIST(Pan.article_ctree), 0, 400);

	gtk_clist_set_column_auto_resize (GTK_CLIST(Pan.article_ctree), 1, TRUE);
	gtk_clist_set_column_justification (GTK_CLIST (Pan.article_ctree), 1, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_auto_resize (GTK_CLIST(Pan.article_ctree), 2, TRUE);
	gtk_clist_set_column_justification (GTK_CLIST (Pan.article_ctree), 2, GTK_JUSTIFY_RIGHT);

/* create the right click popup menu */
	article_ctree_menu = gnome_popup_menu_new (articlelist_menu_popup);

/* connect signals */
	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "button_press_event",
			    GTK_SIGNAL_FUNC(articlelist_button_press), NULL);

	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "click_column",
			    GTK_SIGNAL_FUNC(articlelist_click_column), NULL);

	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "destroy",
	                    GTK_SIGNAL_FUNC(article_ctree_destroy_cb), NULL);


	return s_window;
}

/*FIXME: is this used? */
//static article_db active_group_header_db = NULL;

void
articlelist_set_current_group ( server_data *s, group_data* g )
{
	if (g == my_group)
		return;

	/* OUT WITH THE OLD GROUP */

	if (my_group != NULL)
	{
		ahdb_unref (active_group_header_db);
		active_group_header_db = NULL;
	}

	/* BRING IN THE NEW GROUP */

	my_group = g;
	my_server = s;
	if (g != NULL)
	{
		/* update the title */
		gui_set_title ( g );

		/* open the article header database */
		active_group_header_db = ahdb_ref (s, g);

		/* load the local headers... */
		if ( ahdb_exists ( s, g ) )
			queue_add(QUEUE_ITEM(
				queue_item_headers_new(s, 1, g, HEADERS_LOCAL)));
		else
			choose_download_dialog (s, g);

	}
	else
	{
		articlelist_clear ( );
	}

}
