/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#if defined(__FreeBSD__) || defined(__svr4__)
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#endif

#include <arpa/inet.h>
#include <netdb.h>
#include <pwd.h>
#include "pan.h"

#include "../config.h"

#ifdef AF_NEEDS_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#include "grouplist.h"
#include "nntp.h"
#include "queue.h"
#include "queue-item-send.h"


static const gchar* get_host_name (void);
static const gchar* get_fqdn (const gchar *host);
static gchar* generate_references (Message *msg);

#if 0
static gint
check_from_syntax (gchar *s)
{
	return 1;
}
#endif


#if 0
static gint
check_newsgroups_syntax (gchar *s)
{
	return 1;
}
#endif

typedef enum
{
	OKAY,
	WARN,
	REFUSE,
}
GoodnessLevel;

static void
check_body (
	GSList** setme_errors,
	GoodnessLevel *setme_goodness,
	const gchar* body)
{
	gchar **sd = NULL;
	int i = 0;
	int lines = 0;
	int unquoted_lines = 0;

	g_assert (setme_goodness!=NULL);
	g_assert (setme_errors!=NULL);

	/* check for an empty message */
	if (!body || !*body) {
		*setme_errors = g_slist_append (*setme_errors,
			g_strdup(_("ERROR: The message is empty.")));
		*setme_goodness = MAX (*setme_goodness, REFUSE);
	}

	/* check number of new lines compared to quoted lines */
	unquoted_lines = lines = 0;
	sd = g_strsplit (body, "\n", -1);
	for (i=0; sd[i]!=NULL; ++i) {
		++lines;
		if (*sd[i]!='>')
			++unquoted_lines;
	}
	g_strfreev (sd);
	if (lines && !unquoted_lines) {
		*setme_errors = g_slist_append (*setme_errors,
			g_strdup(_("ERROR: The message is entirely quoted text.")));
		*setme_goodness = MAX (*setme_goodness, REFUSE);
	}
	else if (lines && ((unquoted_lines/(double)lines)<0.25)) {
		*setme_errors = g_slist_append (*setme_errors,
			g_strdup (_("WARNING: The message is mostly quoted text.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}


static gchar *
generate_references (Message *msg)
{
#define GNKSA_LENGTH 998
	gchar buf[GNKSA_LENGTH + 1];
	gchar *p = NULL;
	gchar *refs = NULL;
	gchar *new_refs = NULL;
	gchar *pch = NULL;
	unsigned int total_len, len;

//	p = extract_header (mw->article, "References");
//	if (p)
//		refs = g_strconcat (p, " ", extract_header(mw->article, "Message-ID"), NULL);
//	else
//		refs = g_strconcat (extract_header(mw->article, "Message-ID"), NULL);
//	g_free (p);
	
	if (msg->references) {
		refs = g_strconcat (msg->references, " ", Pan.active_message_id, NULL);
	}
	else if (Pan.active_message_id) {
		refs = g_strconcat (Pan.active_message_id, NULL);
	}
	else {
		refs = g_strdup ( "" );
	}



	len = strlen (refs);
	if (len <= GNKSA_LENGTH) {
		return refs;
	}

	/* The reference header is too long so start truncating */
	/* We have to preserve the first message-id */
	p = strchr(refs, '>');
	if (p == NULL)
		return NULL;
	p++;
	len = p - refs;
	if (len >= GNKSA_LENGTH) {
		return refs;
	}
	strncpy(buf, refs, len);
	buf[len] = 0;
	new_refs = g_strconcat(buf, " ", NULL);
	total_len = len + 1;

	refs += len;
	if ((*refs == '\t') || (*refs == ' '))
		refs++;					/* already handled space */

	/* Now use message-ids at end of the line */
	p = refs + strlen(refs);

	while (1) {
		if (*p == '<') {
			if (strlen(p) + total_len > GNKSA_LENGTH) {
				p = strchr(p + 1, '<');
				if (p == NULL)
					p = refs;
				break;
			}
		}
		if (p == refs)
			break;
		p--;
	}

	pch = new_refs;
	new_refs = g_strconcat (new_refs, p, NULL);
	g_free ( pch );
	g_free ( refs );
	return new_refs;
}

#if 0
static gchar*
generate_message_id (void)
{
	unsigned long now;
	unsigned long randy;
	static unsigned char baseid[64];
	gchar *b, *t, tmp[32];
	char *chars32 = "0123456789abcdefghijklmnopqrstuv";
	static unsigned long last_now;

	while (1) {
		time((time_t *) & now);
		if (now != last_now)
			break;
		sleep (1);
	}
	last_now = now;
	srand (now + getpid());
	randy = (unsigned long) rand();

	b = baseid;
	t = tmp;
	while (now) {
		*t++ = chars32[now & 0x1F];
		now = now >> 5;
	}

	while (t > tmp) {
		t--;
		*b++ = *t;
	}
	*b++ = '.';

	t = tmp;
	while (randy) {
		*t++ = chars32[randy & 0x1F];
		randy = randy >> 5;
	}
	while (t > tmp) {
		t--;
		*b++ = *t;
	}

	*b++ = '@';
	return (char *) baseid;
}
#endif


/* Returns the computers host name */
static const gchar*
get_host_name (void)
{
	gchar *ptr;
	static gchar hostname[256];

	hostname[0] = '\0';

	if ((gethostname(hostname, sizeof(hostname))) == -1) {
		if ((ptr = getenv("HOST")) != NULL)
			strncpy(hostname, ptr, MAXHOSTNAMELEN);
		else if ((ptr = getenv("HOSTNAME")) != NULL)
			strncpy(hostname, ptr, MAXHOSTNAMELEN);
		else
			hostname[0] = '\0';
	}
	hostname[MAXHOSTNAMELEN] = '\0';
	ptr = strtok (hostname, ".");

	return g_strdup ( ptr );
}


/* Returns the fully qualified domain name */
static const gchar *
get_fqdn (const gchar *host)
{
	static char name[512];
	char line[1025];
	char *cp, *domain;
	static char fqdn[1024];
	struct hostent *hp;
	struct in_addr in;
	FILE *inf;

	*fqdn = '\0';
	domain = NULL;

	name[MAXHOSTNAMELEN] = '\0';
	if (host) {
		if (strchr(host, '.'))
			return host;
		strncpy(name, host, MAXHOSTNAMELEN);
	}
	else {
		if (gethostname(name, MAXHOSTNAMELEN))
			return NULL;
	}

	if ('0' <= *name && *name <= '9') {
		in.s_addr = inet_addr(name);
		if ((hp = gethostbyaddr((char *) &in.s_addr, 4, AF_INET)))
			in.s_addr = (*hp->h_addr);
		return (hp && strchr(hp->h_name, '.') ? hp->h_name : (gchar *) inet_ntoa(in));
	}
	if ((hp = gethostbyname(name)) && !strchr(hp->h_name, '.'))
		if ((hp = gethostbyaddr(hp->h_addr, hp->h_length, hp->h_addrtype)))
			in.s_addr = (*hp->h_addr);

	sprintf(fqdn, "%s", hp ? strchr(hp->h_name, '.')
		? hp->h_name : (gchar *) inet_ntoa(in)
		: "");
	if (!*fqdn || (fqdn[strlen(fqdn) - 1] <= '9')) {
		*fqdn = '\0';
		inf = fopen("/etc/resolv.conf", "r");
		if (inf) {
			while (fgets(line, 1024, inf)) {
				line[1024] = '\0';
				g_strchug(line);
				g_strchomp(line);
				if (strncmp(line, "domain ", 7) == 0) {
					domain = line + 7;
					break;
				}
				if (strncmp(line, "search ", 7) == 0) {
					domain = line + 7;
					cp = strchr(domain, ' ');
					if (cp)
						*cp = '\0';
					break;
				}
			}
			if (domain)
				sprintf(fqdn, "%s.%s", name, domain);
			fclose(inf);
		}
	}
	return (fqdn);
}

#if 0
static gchar *
reformat_body (gchar *body)
{
	gchar *tmp;
	gchar *body2 = g_malloc0 (sizeof (body));

	while ((tmp = strtok (body, "\n"))) {
		tmp = wrap_long_lines (tmp, 78);
		g_strconcat (body2, tmp, NULL);
	}

	return body2;
}
#endif

static void
check_mail (
	GSList** setme_errors,
	GoodnessLevel *setme_goodness,
	const GSList* mail_addresses)
{
	const GSList *l;

	for (l=mail_addresses; l!=NULL; l=l->next)
	{
		const gchar* address = (const gchar*) l->data;
		const gchar* at = strchr (address, '@');

		/* find an '@', make sure there's something to the left & right of it */
		if (!at || at==address || !at[1]) {
			*setme_errors = g_slist_append (*setme_errors,
				g_strdup_printf (_("Invalid address: ``%s''"), address));
			*setme_goodness = MAX (*setme_goodness, WARN);
		}

		/* FIXME: validate hostname? */
	}
}


static gint
gslist_compare_group_and_name (gconstpointer a, gconstpointer b)
{
	return g_strcasecmp(((const group_data*)a)->name, (const char*)b);
}

static void
check_groups (
	GSList** setme_errors,
	GoodnessLevel *setme_goodness,
	const server_data* server,
	const GSList* group_names,
	gboolean followup_to_set)
{
	const GSList* l = NULL;
	const int len = g_slist_length ((GSList*)group_names);

	g_return_if_fail (server!=NULL);

	for (l=group_names; l!=NULL; l=l->next) {
		if (!g_slist_find_custom (server->grouplist, l->data, gslist_compare_group_and_name)) {
			gchar* warn = g_strdup_printf (_("ERROR: Unknown group: ``%s''."), (char*)l->data);
			*setme_errors = g_slist_append (*setme_errors, warn);
			*setme_goodness = MAX (*setme_goodness, REFUSE);
		}
	}

	if (len>5) {
		*setme_errors = g_slist_append (*setme_errors,
			g_strdup(_("WARNING: Posting to a large number of groups.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}

	if (len>2 && !followup_to_set) {
		*setme_errors = g_slist_append (*setme_errors,
			g_strdup(_("WARNING: Crossposting to a number of groups without setting Followup-To.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}



/* Find and return a list of e-mail addresses to send to */
static GSList *
get_email_rcpts (gchar *to)
{
	gchar *dest;
	GSList *email_rcpts = NULL;

	g_return_val_if_fail (to!=NULL, NULL);

	dest = strtok (to, ",");

	while (dest != NULL)
	{
		while (*dest == ' ') dest++; /* skip over whitespaces */

		if (strchr (dest, '@'))
			email_rcpts = g_slist_append ( email_rcpts, dest );

		dest = strtok (NULL, ",");
	}

	return email_rcpts;
}


/* Find and return a list of NNTP groups to send to */
static GSList*
get_nntp_rcpts (const gchar *to)
{
	gchar *pch = g_strstrip(g_strdup(to));
	GSList *nntp_rcpts = NULL;
	gchar ** split = NULL;
	int i = 0;

	/* check simple cases */
	if (!to || !*to) {
		g_free (pch);
		return NULL;
	}

	/* split the list */
	pch = g_strdelimit (pch, " ;:", ',');
	split = g_strsplit (pch, ",", -1);
	g_free (pch);

	/* walk through the list */
	for (i=0; split[i]!=NULL; ++i)
		nntp_rcpts = g_slist_append (nntp_rcpts, g_strstrip(split[i]));

	/* cleanup */
	g_free (split);

	return nntp_rcpts;
}

static GSList*
get_smtp_rcpts (const gchar *to)
{
	/*FIXME: handle addresses w/spaces "Joe Bob <joebob@devnull.com>" */

	return get_nntp_rcpts (to);
}

static gchar*
build_nntp_message (Message *msg)
{
	gchar *subject = NULL;
	gchar *from = NULL;
	gchar *newsgroups = NULL;
	gchar *message = NULL;
	gchar *path = NULL;
	//gchar *organization = NULL;
	gchar *body = NULL;
	gchar *buf = NULL;
	static const char *host = NULL;
	static const char *fqdn = NULL;
	static char *uname = NULL;
	//gchar *err;
	GSList *l = NULL;
	gchar *tmp = NULL;

	g_return_val_if_fail (msg!=NULL, NULL);

	if (host == NULL)
		host = get_host_name ();
	if (fqdn == NULL)
		fqdn = get_fqdn (host);
	if (uname == NULL)
		uname = g_get_user_name ();

	path = g_strconcat (host, "!", uname, NULL);

	for (l=msg->rcpts; l!=NULL; l=l->next)
		newsgroups = pan_append (newsgroups, (gchar*)l->data, (newsgroups?',':'\0'));

	from = gtk_editable_get_chars (GTK_EDITABLE (msg->window->from), 0, -1);
	subject = gtk_editable_get_chars (GTK_EDITABLE (msg->window->subject), 0, -1);
	//organization = gtk_editable_get_chars (GTK_EDITABLE (msg->window->organization), 0, -1);
	body = gtk_editable_get_chars (GTK_EDITABLE (msg->window->body), 0, -1);

#if 0
	err = check_body ( body );
	if ( err != NULL ) {
		pan_error_dialog ( _("Unable to post message:\n%s"), err );
		g_free ( body );
		return NULL;
	}
	g_free ( body );
#endif

/* FIXME: doesn't do anything
	if (!check_from_syntax(from))
		g_error ("Bad From. Do something!");
	if (!check_newsgroups_syntax(newsgroups))
		g_error ("Bad Newsgroups. Do something!");
*/
	message = g_strconcat ("From: ", from, "\n",
			       "Path: ", path, "\n",
			       "Newsgroups: ", newsgroups, "\n",
			       "Subject: ", subject, "\n", NULL);

	g_free (path);
	g_free (from);
	g_free (newsgroups);
	g_free (subject);
	tmp = message;

	/* FIXME
	if (Pan.active_server->gen_msgid) {
		message = g_strconcat (tmp,"Message-ID: <", generate_message_id(), fqdn, ">\n", NULL);
		g_free (tmp);
		tmp = message;
	}
	*/
	if (msg && msg->type==NNTP_ARTICLE_REPLY) {
		gchar* pch = generate_references (msg);
		message = g_strconcat (message, "References: ", pch, "\n", NULL);
		g_free (pch);
		g_free (tmp);
		tmp = message;
	}
	body = g_strconcat ("\n",
		gtk_editable_get_chars(GTK_EDITABLE(msg->window->body), 0, -1),
		NULL);
	buf = generate_date ();
	message = g_strconcat (tmp,
			       "Date: ", buf, "\n",
			       //"Organization: ", organization, "\n",
			       "X-Newsreader: ", "Pan ", VERSION, "\n",
			       body, NULL);
	g_free (buf);
	g_free (tmp);
	g_free (body);

	return message;
}

static void
check_subject (
	GSList** setme_errors,
	GoodnessLevel *setme_goodness,
	const gchar* subject)
{
	g_assert (setme_errors!=NULL);
	g_assert (setme_goodness!=NULL);

	if (!subject || !*subject) {
		*setme_errors = g_slist_append (*setme_errors,
			g_strdup(_("ERROR: No subject specified.")));
		*setme_goodness = MAX (*setme_goodness, REFUSE);
	}
}


static int
send_done_cb (gpointer call_object, gpointer call_arg, gpointer user_data)
{
	StatusItem* item = STATUS_ITEM(call_object);
	int status = GPOINTER_TO_INT(call_arg);
	GtkWidget* send_button = GTK_WIDGET(user_data);
	Message *message = QUEUE_ITEM_SEND(item)->msg;
	GSList *lines = NULL;
	GSList *l = NULL;
	gchar *text = NULL;

	if (status & SMTP_FAILED)
		lines = g_slist_append (lines, _("E-Mail send failed."));
	if (status & NNTP_FAILED)
		lines = g_slist_append (lines, _("Usenet posting failed."));
	if (!status)
		lines = g_slist_append (lines, _("Message sent successfully."));

	for (l=lines; l!=NULL; l=l->next)
		text = pan_append (text, (const gchar*)l->data, '\n');

	pan_lock();
	gnome_ok_dialog_parented (text, GTK_WINDOW(message->window->window));
	g_free (text);

	if (!status)
	{
		gtk_widget_destroy (message->window->window);
		message_free (message);
	}
	else
	{
		gtk_widget_set_sensitive (send_button, TRUE);
	}
	pan_unlock();

	return 0;
}

static void
start_sending (GtkWidget *widget, server_data* sdata, Message *msg)
{
	PanObject *item = queue_item_send_new (sdata, msg);
	gtk_widget_set_sensitive (GTK_WIDGET(widget), FALSE);
	pan_callback_add (STATUS_ITEM(item)->done, send_done_cb, widget);
	queue_add (QUEUE_ITEM(item));
}

typedef struct {
	GtkWidget *widget;
	server_data *sdata;
	Message *msg;
} SendAnywayCallbackStruct;

static void
send_anyway_cb (gint reply, gpointer data)
{
	if (!reply) { // yes
		SendAnywayCallbackStruct *cbs = (SendAnywayCallbackStruct*)data;
		start_sending (cbs->widget, cbs->sdata, cbs->msg);
	}

	g_free (data);
}

void
send_cb (GtkWidget *widget, Message *msg)
{
	gchar *subject = NULL;
	gchar *body = NULL;
	GoodnessLevel goodness = OKAY;
	GSList *errors = NULL;
	GSList *groups_slist = NULL;
	GSList *mail_slist = NULL;
	server_data *server = grouplist_get_current_server();
	int group_qty = 0;
	int mail_qty = 0;
	int error_len = 0;
	gboolean followup_to_set;

	/* sanity check */
	g_return_if_fail (widget!=NULL);
	g_return_if_fail (msg!=NULL);
	g_return_if_fail (msg->window!=NULL);
	g_return_if_fail (msg->window->subject!=NULL);

	/* check the subject... */
	subject = gtk_editable_get_chars (GTK_EDITABLE (msg->window->subject), 0, -1);
	subject = g_strstrip (subject);
	check_subject (&errors, &goodness, subject);
	g_free (subject);

	/* check the body... */
	body = gtk_editable_get_chars (GTK_EDITABLE (msg->window->body), 0, -1);
	body = g_strstrip (body);
	check_body (&errors, &goodness, body);
	g_free (body);

	/* FIXME: check followup-to */
	followup_to_set = FALSE;

	/* check the groups... */
	groups_slist = get_nntp_rcpts(gtk_entry_get_text(GTK_ENTRY(msg->window->newsgroups)));
	group_qty = g_slist_length (groups_slist);
	check_groups (&errors, &goodness, server, groups_slist, followup_to_set);
	g_slist_foreach (groups_slist, (GFunc)g_free, NULL);
	g_slist_free (groups_slist);

	/* FIXME: check "from" */

	/* check the email recipients... */
	mail_slist = get_smtp_rcpts(gtk_entry_get_text(GTK_ENTRY(msg->window->users)));
	mail_qty = g_slist_length (mail_slist);
	check_mail (&errors, &goodness, mail_slist);
	g_slist_foreach (mail_slist, (GFunc)g_free, NULL);
	g_slist_free (mail_slist);

	/* one last error check */
	if (!group_qty && !mail_qty) {
		errors = g_slist_append (errors, g_strdup(_("ERROR: No recipients specified.")));
		goodness = MAX (goodness, REFUSE);
	}

	error_len = g_slist_length (errors);
	if (error_len)
	{
		/* post a warning dialog */
		GSList *l = NULL;
		gchar *pch = NULL;
		for (l=errors; l!=NULL; l=l->next)
			pch = pan_append (pch, (const gchar*)l->data, '\n');
		if (goodness==REFUSE)
			pan_error_dialog (pch);
		else {
			SendAnywayCallbackStruct* cbs = g_new0 (SendAnywayCallbackStruct, 1);
			const gchar *ques = error_len>1
				? _("Do you want to post anyway, despite these warnings?")
				: _("Do you want to post anyway, despite this warning?");
			pch = pan_append (pch, "\n\n", 0);
			pch = pan_append (pch, ques, 0);
			cbs->widget = widget;
			cbs->sdata = server;
			cbs->msg = msg;
			gnome_question_dialog (pch, send_anyway_cb, cbs);
		}
		g_warning (pch);
		g_free (pch);

		/* cleanup */
		g_slist_foreach (errors, (GFunc)g_free, NULL);
		g_slist_free (errors);
		errors = NULL;
	}

	if (goodness==OKAY)
		start_sending (widget, server, msg);
}

gint
message_send (StatusItem *item, PanSocket *sock, Message *msg)
{
	GSList *nntp_rcpts;
	GSList *email_rcpts;
	gchar *nntp_msg;
	gchar *tmp1, *tmp2;
	gchar *newsgroups;
	gchar *users;
	gchar *subject;
	gint retval = 0;

	g_return_val_if_fail (msg!=NULL, ~0);

	subject = gtk_editable_get_chars (
		GTK_EDITABLE (msg->window->subject), 0, -1);

	/* where is this being posted... */
	newsgroups = gtk_editable_get_chars (
		GTK_EDITABLE (msg->window->newsgroups), 0, -1);
	tmp1 = g_strdup (newsgroups);
	nntp_rcpts = get_nntp_rcpts (tmp1);

	/* who is this being mailed to... */
	users = gtk_editable_get_chars (
		GTK_EDITABLE (msg->window->users), 0, -1);
	tmp2 = g_strdup (users);
	email_rcpts = get_email_rcpts (tmp2);


	if (email_rcpts) {
		msg->rcpts = email_rcpts;
		if (!smtp_send (item, msg))
			log_add_va (_("E-mail Message '%s' sent."), subject);
		else {
			log_add_va (_("E-mail Message '%s' not sent."),subject);
			retval |= SMTP_FAILED;
		}
	}


	if (nntp_rcpts) {
		msg->rcpts = nntp_rcpts;
		nntp_msg = build_nntp_message (msg);

		if (!nntp_post (sock, nntp_msg))
			log_add_va (_("NNTP Article '%s' posted."), subject);
		else
			retval |= NNTP_FAILED;

		g_free (nntp_msg);
	}


	/* clean up */
	g_free (subject);
	g_free (tmp1);
	g_free (tmp2);
	g_free (newsgroups);
	g_free (users);

	return retval;
}
